{*******************************************************************************
This is a program, which computes in real time the spectrum of an audio signal,
using the Fast Fourier Transform (FFT-Radix-2). The output can be viewed on an
conventional oscilloscope, connected at the outputs of the DAC-8412. In this 
case the computed spectrum of the signal appears on DACA-output, and the 
SYNC-pulses appear on DACD-output. An average view of the spectrum is available 
on DACB. 
*******************************************************************************}

.module/RAM/ABS=0 fft_program;

{*************** Constant Declarations ***************}
{*****memory mapped ADSP-2181 control registers*******}
.const IDMA=				0x3fe0;
.const BDMA_BIAD=			0x3fe1;
.const BDMA_BEAD=			0x3fe2;
.const BDMA_Ctrl=			0x3fe3;
.const BDMA_BWCOUNT=			0x3fe4;
.const PFDATA=				0x3fe5;
.const PFTYPE=				0x3fe6;
.const SPORT1_Autobuf=			0x3fef;
.const SPORT1_RFSDIV=			0x3ff0;
.const SPORT1_SCLKDIV=			0x3ff1;
.const SPORT1_Control_Reg=		0x3ff2;
.const SPORT0_Autobuf=			0x3ff3;
.const SPORT0_RFSDIV=			0x3ff4;
.const SPORT0_SCLKDIV=			0x3ff5;
.const SPORT0_Control_Reg=		0x3ff6;
.const SPORT0_TX_Channels0=		0x3ff7;
.const SPORT0_TX_Channels1=		0x3ff8;
.const SPORT0_RX_Channels0=		0x3ff9;
.const SPORT0_RX_Channels1=		0x3ffa;
.const TSCALE=				0x3ffb;
.const TCOUNT=				0x3ffc;
.const TPERIOD=				0x3ffd;
.const DM_Wait_Reg=			0x3ffe;
.const System_Control_Reg=		0x3fff;
.const N=				1024; {number of FFT-points}
.const Ntimes2 = 			2048; {two buffers together}
.const N_div_2 = 			512;
.const mod_value = 			H#0010; {modifier for reversing N bits}
.const log2N = 				10;	{number of stages of the FFT  } 
.const nover2 = 			512;
.const nover4 = 			256;
.const DACA   =				0;	{IO-addresses of the DAC-registers}
.const DACB   = 			128;	{of DAC-8412}
.const DACC   =				256;
.const DACD   =				384;
.const DACresol =			4096;	{this corresponds to 12-bit}

{***** Variable and Buffer Declarations *****************}
.var/pm/ram/circ	 twid_real[N_div_2];
.var/pm/ram/circ	 twid_imag[N_div_2];
.var/pm/ram/circ	 hanning[N];
.var/dm/ram/circ/abs=0	 inplacereal[N], inplaceimag[N], padding[4];
.var/dm/ram/circ	 inputreal[N], inputimag[N];
.var/dm/ram/circ	 magnitude[N], averageval[N], squareval[DACresol];
.var/dm/ram	 groups, bflys_per_group, node_space, blk_exponent, blockexp;
.var/dm/ram/circ rx_buf[3];	{ AD1847 receive buffer }
.var/dm/ram/circ tx_buf[3];	{ AD1847 transmit buffer }
.var/dm/ram/circ init_cmds[13];
.var/dm/ram	 monflag;
.var/dm/ram      stat_flag;
.var/dm/ram	 middlep;
.var/dm/ram 	 datablock;	{block of ready data to be processed}
.var/dm/ram	 passed;	{FFT passed flag}

{***** Variable and buffer initialization ***************}
.init 	twid_real: <tw_real.dat>;
.init	twid_imag: <tw_imag.dat>;
.init 	hanning: <hanning.dat>;
.init 	inputreal: <probe.dat>;
.init   inputimag: <probe.dat>;
.init 	inplaceimag: <probe.dat>;
.init 	magnitude: <probe.dat>;
.init   squareval: <squarval.dat>;
.init 	groups: N_div_2;
.init 	bflys_per_group: 2;
.init	node_space: 2;
.init 	blk_exponent: 0;
.init 	blockexp: 0;
.init rx_buf: 	0x0001, 0x0001, 0x0001;
.init tx_buf:	0xc000, 0x0000, 0x0000;
.init init_cmds:
	0xc003,		{AD1847 Left input control register}
	{0 0 0 0 0 0 0 0
	 | | | | |_|_|_|_Left Input Gain Select, LSB represents +1.5dB
	 | | |_|_________reserved
	 |_|_____________Left Input Source Select
			 	0 - Left Line 1
				1 - Left Auxiliary 1
				2 - Left Line 2
				3 - Left Line 1 Post-Mixed Output Loopback
	}
	0xc103,		{AD1847 Right input control register}
	{0 0 0 0 0 0 0 0
	 | | | | |_|_|_|_Right Input Gain Select, LSB represents +1.5dB
	 | | |_|_________reserved
	 |_|_____________Right Input Source Select
			 	0 - Right Line 1
				1 - Right Auxiliary 1
				2 - Right Line 2
				3 - Right Line 1 Post-Mixed Output Loopback
	}
	0xc288,		{AD1847 Left aux1 control register}
	{1 0 0 0 0 0 0 0
	 | | | |_|_|_|_|_Left Auxiliary Input 1 Gain Select,0000=+12dB,LSB=-1.5dB
	 | |_|___________reserved
	 |_______________Left Auxiliary 1 Mute
	}
	0xc388,		{AD1847 Right aux1 control register}
	{1 0 0 0 0 0 0 0
	 | | | |_|_|_|_|_Right Auxiliary Input 1 Gain Select,0000=+12dB,LSB=-1.5dB
	 | |_|___________reserved
	 |_______________Right Auxiliary 1 Mute
	}
	0xc488,		{AD1847 Left aux2 control register}
	{1 0 0 0 0 0 0 0
	 | | | |_|_|_|_|_Left Auxiliary Input 2 Gain Select,0000=+12dB,LSB=-1.5dB
	 | |_|___________reserved
	 |_______________Left Auxiliary 2 Mute
	}
	0xc588,		{AD1847 Right aux2 control register}
	{1 0 0 0 0 0 0 0
	 | | | |_|_|_|_|_Right Auxiliary Input 2 Gain Select,0000=+12dB,LSB=-1.5dB
	 | |_|___________reserved
	 |_______________Right Auxiliary 2 Mute
	}
	0xc680,		{AD1847 Left DAC control register}
	{1 0 0 0 0 0 0 0
	 | | |_|_|_|_|_|_Left DAC Attenuate Select,000000=0dB,LSB=-1.5dB
	 | |_____________reserved
	 |_______________Left DAC Mute if HI
	}
	0xc780,		{AD1847 Right DAC control register}
	{1 0 0 0 0 0 0 0
	 | | |_|_|_|_|_|_Right DAC Attenuate Select,000000=0dB,LSB=-1.5dB
	 | |_____________reserved
	 |_______________Right DAC Mute if HI
	}
	0xc85b,		{AD1847 data format register}
	{0 0 0 0 0 0 0 0
	 | | | | | | | |_Clock Source Select,0=24.576MHz,1=16.9344MHz
	 | | | | |_|_|___Clock Frequency Divide Select, XTAL= 24.576    16.9344
	 | | | |					0      8.0	 5.5125
	 | | | |					1     16.0	11.025 
	 | | | |		in kHz			2     27.42857  18.9
	 | | | |					3     32.0      22.05  
	 | | | |					4     not supp. 37.8
	 | | | |					5     not supp. 44.1
	 | | | |					6     48.0      33.075
	 | | | |					7      9.6       6.615
	 | | | |_________Stereo/Mono Select 0-Mono 1-Stereo
	 | | |___________Companded/Linear Select 0-Linear PCM 1-Companded
	 | |_____________Format Select 0 = 8-bit unsigned linear PCM   8-bit mu-law
	 |			       1 = 16-bit signed linear PCM    8-bit A-law 	
	 |_______________reserved
	}
	0xc909,		{AD1847 interface configuration register}
	{0 0 0 0 1 0 0 0
	 | | | | | | | |_Playback Enable 0=disabled 1=enabled
	 | | | | | |_|___reserved
	 | | | | |_______Autocalibrate enable 0=No autocalibration 1=Autocalibration
	 |_|_|_|_________reserved
	}
	0xca00,		{AD1847 pin control register}
	{0 0 0 0 0 0 0 0
	 | | | |_|_|_|_|_reserved
	 | | |___________Clockout Three-State, 1=CLKOUT pin is three-stated, if BM=high
	 |_|_____________External Control   State=XCTL1,XCTL0 pins  
	} 
	0xcc40,		{AD1847 miscellaneous information register}
	{0 0 0 0 0 0 0 0
	 | | |_|_|_|_|_|_reserved
	 | |_____________Transmit Slot Select 0=Transmit on time slots 3,4,5 "1-wire system"
         |                                    1=Transmit on time slots 0,1,2 "2-wire system" 
	 |_______________Frame Size 0= 32 slots per frame, 1= 16 slots per frame 
	}
	0xcd00;		{AD1847 digital mix control register}
	{0 0 0 0 0 0 0 0
	 | | | | | | | |_Digital Mix Enabled (0=disabled 1=enabled) ADCout+DACin 
	 | | | | | | |___reserved
	 |_|_|_|_|_|_____Digital Mix Attenuation of ADCout (LSB=-1.5dB)
	}
.init 	monflag:	0x0000;
.init 	passed: 1;	{begins with collecting data, not with a FFT}

{********** Interrupt Vector Table *************}
       jump start; 		{Location 0000: reset}
	rti;
	rti;
	rti;

	rti; 			{Location 0004: IRQ2}
	rti;
	rti;
	rti;

	rti;			{Location 0008: IRQL1}
	rti;
	rti;
	rti;

	rti;			{Location 000C: IRQL0}
	rti;
	rti;
	rti;

	ar = dm(stat_flag);	{Location 0010: SPORT0 tx}
	ar = pass ar;
	if eq rti;
	jump next_cmd;

	jump input_samples;	{Location 0014: SPORT0 rx}
	rti;
	rti;
	rti;

	ax0 = 1;		{Location 0018: IRQE}
	dm(monflag) = ax0;
	rti;
	rti;

	rti;			{Location 001C: BDMA}
	rti;
	rti;
	rti;

	rti;			{Location 0020: SPORT1 tx or IRQ1}
	rti;
	rti;
	rti;

	rti;			{Location 0024: SPORT1 rx or IRQ0}
	rti;
	rti;
	rti;

	rti;			{Location 0028: timer}
	rti;
	rti;
	rti;

	rti;			{Location 002C: power down}
	rti;
	rti;
	rti;

{************* ADSP 2181 initialization ****************}
start:	i6 = ^rx_buf;	{load address pointer with buffer start}
	l6 = %rx_buf;	{load length register with buffer size}
	i7 = ^tx_buf;
	l7 = %tx_buf;
	i3 = ^init_cmds;
	l3 = %init_cmds;
	m1 = 1;
	m7 = 1;

{************* Serial Port 0 (SPORT0) Set Up *************}
	{SPORT0 Autobuffer Control Register
	 0 0 0 0 X X X X X X X X X X 0 0
	 | | | | | | | | | | | | | | | |_Receive Autobuffering Enable
	 | | | | | | | | | | | | | | |___Transmit Autobuffering Enable
	 | | | | | | | | | | | | |_|_____RMREG autobuffer receive modifier
	 | | | | | | | | | |_|_|_________RIREG autobuffer receive index reg.
	 | | | | | | | |_|_______________TMREG autobuffer transmit modifier
	 | | | | |_|_|___________________TIREG autobuffer transmit index reg.
	 | | | |_________________________MAC Biased Rounding Control Bit
	 | | |___________________________Reserved Bit
	 | |_____________________________CLKOUT Disable Control Bit
	 |_______________________________Reserved Bit
	} 
	ax0 = b#0000111111101111;
	dm (SPORT0_Autobuf) = ax0;
	ax0 = 0;
	{SPORT0_RFSDIV=Receive Frame Sync Divide Modulus
	 RFSDIV = (SCLK frequency/RFS frequency)-1
	}
	dm (SPORT0_RFSDIV) = ax0;
	{SPORT0_SCLKDIV=Serial Clock Divide Modulus
	 SCLKDIV = (CLKOUT frequency/2*SCLK frequency)-1
	}
	dm (SPORT0_SCLKDIV) = ax0;
	{
	ax0 = b#0000000000000111;
	dm (SPORT0_TX_Channels0) = ax0;
	ax0 = b#0000000000000111;
	dm (SPORT0_TX_Channels1) = ax0;
	ax0 = b#0000000000000111;
	dm (SPORT0_RX_Channels0) = ax0;
	ax0 = b#0000000000000111;
	dm (SPORT0_RX_Channels1) = ax0;

{************* Serial Port 1 (SPORT1) Set Up **************}
	ax0 = 0;
	{SPORT1 Autobuffer Control Register
	 0 0 0 0 X X X X X X X X X X 0 0
	 | | | | | | | | | | | | | | | |_Receive Autobuffer Enable
	 | | | | | | | | | | | | | | |___Transmit Autobuffer Enable
	 | | | | | | | | | | | | |_|_____autobuffer receive modifier
	 | | | | | | | | | |_|_|_________autobuffer receive index register
	 | | | | | | | |_|_______________autobuffer transmit modifier
	 | | | | |_|_|___________________autobuffer transmit index register
	 | | | |_________________________Powerup Context Reset 1=soft reset
	 | | |___________________________Powerdown interrupt force
	 | |_____________________________4096 Cycl.delay to Startup from PWRDown
	 |_______________________________XTAL-pin disable during Powerdown
	}
	dm (SPORT1_Autobuf) = ax0;	{autobuffering disabled}
	dm (SPORT1_RFSDIV) = ax0;	{RFSDIV not used}
	dm (SPORT1_SCLKDIV) = ax0;	{SCLKDIV not used}
	{SPORT1 Control Register
	 X 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 
	 | | | | | | | | | | | | |_|_|_|_Serial Word Length - 1
	 | | | | | | | | | | |_|_________data format 00-right just.,zero fill
         | | | | | | | | | |                         01-right just.,sign-extend
         | | | | | | | | | |                         10-compand mu-law
	 | | | | | | | | | |	                     11-compand A-law			
	 | | | | | | | | | |_____________Invert Receive Frame Sync
	 | | | | | | | | |_______________Invert Transmit Frame Sync
	 | | | | | | | |_________________Internal Receive Frame Sync Enable
	 | | | | | | |___________________Internal Transmit Frame Sync Enable
	 | | | | | |_____________________Transmit Frame Sync Width
	 | | | | |_______________________Transmit Frame Sync Required
	 | | | |_________________________Receive Frame Sync Width
	 | | |___________________________Receive Frame Sync Required
	 | |_____________________________Internal Serial Clock Generation
	 |_______________________________Flag Out (read-only)
	} 
	dm (SPORT1_Control_Reg) = ax0;	{control disabled}


{************* Timer Set Up *******************************}
	ax0=0;
	dm(TSCALE) = ax0;		{timer not used}
	dm(TCOUNT) = ax0;
	dm(TPERIOD) = ax0;

{************* System and Memory Set Up *******************}
	{Waitstate Control Register
	 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
	 | | | | | | | | | | | | | |_|_|_IOWAIT0
	 | | | | | | | | | | |_|_|_______IOWAIT1
	 | | | | | | | |_|_|_____________IOWAIT2
	 | | | | |_|_|___________________IOWAIT3
	 | |_|_|_________________________DWAIT
	 |_______________________________1=ROM enabled 0=ROM disabled
	}
	ax0 = b#0000111111111111;
	{ax0 = b#0000000000000000;}
	dm (DM_Wait_Reg) = ax0;
	{System Control Register
	 0 0 0 0 0 1 0 0 0 0 X 1 1 1 1 1
	 | | | | | | | | | | | | | |_|_|_Program Memory Wait States
	 | | | | | | | | | | |_|_|_______Not used on ADSP-2181				
         | | | | | | | |_|_|_____________Not used on ADSP-2181 
	 | | | | | | |___________________Not used on ADSP-2181
	 | | | | | |_____________________Sport1 Configure 1=serial port 0=FI,FO,IRQ0,IRQ1,SCLK
         | | | | |_______________________Sport1 Enable 1=enabled 0=disabled
         | | | |_________________________Sport0 Enable 1=enabled 0=disabled
	 |_|_|___________________________Reserved bits
	} 		
	ax0 = b#0001000000000000;	{SPORT0 enabled}
	dm (System_Control_Reg) = ax0;
	{Interrupt Force Clear
	 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 
	 | | | | | | | | | | | | | | | |_Timer Interrupt Clear Bit
       	 | | | | | | | | | | | | | | |___Sport1 Receive or IRQ0 Clear Bit
	 | | | | | | | | | | | | | |_____Sport1 Transmit or IRQ1 Clear Bit
	 | | | | | | | | | | | | |_______BDMA Interrupt Clear Bit
	 | | | | | | | | | | | |_________IRQE Clear Bit
	 | | | | | | | | | | |___________Sport0 Receive Clear Bit
	 | | | | | | | | | |_____________Sport0 Transmit Clear Bit
	 | | | | | | | | |_______________IRQ2 Clear Bit 
	 | | | | | | | |_________________Timer Interupt Force Bit
	 | | | | | | |___________________Sport1 Receive or IRQ0 Force Bit
	 | | | | | |_____________________Sport1 Transmit or IRQ1 Force Bit
	 | | | | |_______________________BDMA Interrupt Force Bit
	 | | | |_________________________IRQE Force Bit 
	 | | |___________________________Sport0 Receive Force Bit
	 | |_____________________________Sport0 Transmit Force Bit
	 |_______________________________IRQ2 Force Bit
	}
	ifc = b#00000011111111;		{clear pending interrupts}
	nop;
	{Interrupt Control
	 X X X X X
	 | | | | |_IRQ0 Sensitivity 1=Edge O=Level
	 | | | |___IRQ1 Sensitivity 1=Edge 0=Level
	 | | |_____IRQ2 Sensitivity 1=Edge 0=Level
	 | |_______Reserved Bit
	 |_________Interrupt Nesting 1=Enable 0=Disable
	}
	icntl = b#00000;		{level sensitive interrupts}
	{MSTAT
	 0 0 0 0 0 0 0
	 | | | | | | |_Register Bank Select 0=primary 1=secondary
	 | | | | | |___Bit-Reverse Addressing Enable (DAG1)
	 | | | | |_____ALU Overflow Latch Mode Enable
	 | | | |_______AR Saturation Mode Enable
	 | | |_________MAC Result Placement 0=fractional 1=integer
	 | |___________Timer Enable
	 |_____________Go Mode Enable
	}
	mstat = b#1000000;

{************ AD1847 Codec initialization *****************}
	ax0 = 1;
	dm (stat_flag) = ax0;		{set the status flag}
	{Interrupt Mask IMASK
	 0 0 0 0 0 0 0 0 0 0    1=interrupt enable 0=interrupt disable
	 | | | | | | | | | |_Timer
	 | | | | | | | | |___Sport1 Receive or IRQ0
	 | | | | | | | |_____Sport1 Transmit or IRQ1
	 | | | | | | |_______BDMA Interrupt
	 | | | | | |_________IRQE
	 | | | | |___________Sport0 Receive
	 | | | |_____________Sport0 Transmit
	 | | |_______________IRQL0
	 | |_________________IRQL1
	 |___________________IRQ2
	}
	imask = b#0001000000;		{enable transmit interrupt}
	ax0 = dm(i7,m7);		{start interrupt}
	tx0 = ax0;

check_init:
	ax0 = dm (stat_flag);		{wait for the entire}
	af = pass ax0;			{init.buffer to be  }
	if ne jump check_init;		{sent to the CODEC  }

	ay0 = 2;
check_acih:
	ax0 = dm (rx_buf);		{wait for CODEC to  }
	ar = ax0 and ay0;		{come out of	    }
	if eq jump check_acih;		{autocalibration    }

check_acil:
	ax0 = dm (rx_buf);		{wait for CODEC to  }
	ar = ax0 and ay0;		{come out of 	    }
	if ne jump check_acil;		{autocalibration    }

	idle;

	ay0 = 0xbf3f;			{unmute left DAC}
	ax0 = dm (init_cmds + 6);
	ar = ax0 AND ay0;
	dm (tx_buf) = ar;
	idle;

	ax0 = dm (init_cmds + 7);	{unmute right DAC}
	ar = ax0 AND ay0;
	dm (tx_buf) = ar;
	idle;

	ax0 = 0xc901;			{clear autocalibration request}
	dm (tx_buf) = ax0;
	idle;
	ax1 = 0x8000;			{control word to clear over-range fl.}
	dm (tx_buf) = ax1;
	
	ifc = b#00000011111111;		{clear any pending interrupt}
	nop;

	imask = b#0000110000;		{enable rx0 interrupt}
	

{****** loading the input data samples and FFT-main *********************}
main: 	ax0 = 0x0fff;
	IO(DACD) = ax0;		{SYNC-pulse generation on DACD of DAC8412}
	i0 = ^inputreal;
  	i1 = ^magnitude;
	i3 = ^averageval;
	m0 = 1;
	m1 = 1;
	m3 = 1;
	l0 = %inputreal;
	l1 = %magnitude;
	l3 = %averageval;
	ena ints;

main1:	ax0 = dm(monflag);	{checks, if the INTERRUPT-key is pressed}
	af = pass ax0;		{if pressed, then return to monitor	}
	if ne rts;

	idle;

	ax0 = dm(passed);	{checks, if the FFT is passed once} 
	af = pass ax0;
	if ne jump main1;
	
	dis ints;

	ax0 = 0x0000;
	IO(DACD) = ax0;		{SYNC-pulse generation}
	IO(DACA) = ax0;	
	IO(DACB) = ax0;		{reset the DAC-s at the end of the current sweep}

windowr:i0 = ^inputreal;	{Window for the real input data}
	m0 = 1;
	l0 = %inputreal;
	i2 = ^inputreal;
	m2 = 1;
	l2 = %inputreal;
	i4 = ^hanning;
	m4 = 1;
	l4 = %hanning;
	cntr = N;
	mr = 0;
	mx0 = dm(i0,m0);
	my0 = pm(i4,m4);
	do winr until ce;
		mr = 0;
		mr = mx0*my0 (RND), mx0 = dm(i0,m0), my0 = pm(i4,m4);
	winr:	dm(i2,m2) = mr1;

windowi:i0 = ^inputimag;	{Window for the imaginary data}
	m0 = 1;
	l0 = %inputimag;
	i2 = ^inputimag;
	m2 = 1;
	l2 = %inputimag;
	i4 = ^hanning;
	m4 = 1;
	l4 = %hanning;
	cntr = N;
	mr = 0;
	mx0 = dm(i0,m0);
	my0 = pm(i4,m4);
	do wini until ce;
		mr = 0;
		mr = mx0*my0 (RND), mx0 = dm(i0,m0), my0 = pm(i4,m4);
	wini:	dm(i2,m2) = mr1;	

scramble:
	i4 = ^inputreal;
	m4 = 1;
	i0 = ^inplacereal;	{scrambled data}
	m0 = mod_value;
	l4 = 0;
	l0 = 0;
	i2 = ^inplaceimag;
	m2 = 1;
	CNTR = N;
	ENA BIT_REV;		{enable bit-reversed outputs on DAG1}
	do brev until ce;
		ay1 = dm (i4,m4);
		dm (i0,m0) = ay1;  {writes data in bit-reversed location}
brev:		dm (i2,m2) = 0;	   {zeroes the imaginary data buffer}
	DIS BIT_REV;		{disable bit-reversed mode}

fftstart:
	CNTR = log2N - 2; 	{initialize the FFT stage counter}
	m0 = 0;
	m1 = 1;
	l1 = 0;
	l2 = 0;
	l3 = 0;
	l4 = %twid_real;
	l5 = %twid_imag;
	l0 = 0;
	SB = -2;

{****************** STAGE 1 of the FFT transform *************************}
	i0 = ^inplacereal;
	i1 = ^inplacereal + 1;
	i2 = ^inplaceimag;
	i3 = ^inplaceimag + 1;
	m2 = 2;

	CNTR = nover2;
	ax0 = dm (i0,m0);
	ay0 = dm (i1,m0);
	ay1 = dm (i3,m0);

	do group_lp until ce;
		ar = ax0 + ay0, ax1 = dm (i2,m0);
		SB = expadj ar, dm (i0,m2) = ar;
		ar = ax0 - ay0;
		SB = expadj ar;
		dm (i1,m2) = ar, ar = ax1 + ay1;
		SB = expadj ar, dm (i2,m2) = ar;
		ar = ax1 - ay1, ax0 = dm (i0,m0);
		SB = expadj ar, dm (i3,m2) = ar;
		ay0 = dm (i1,m0);
group_lp:	ay1 = dm (i3,m0);
		call bfp_adj;

{*************** STAGES 2 to log2N-1 of FFT transform *********************}
do stage_loop until ce; 		{computes all stages in FFT}
		i0 = ^inplacereal;	{x0 of the 1-st group}
		i2 = ^inplaceimag;	{y0 of the 1-st group}
		SI = dm (groups);
		SR = ashift SI by -1 (lo); {groups = groups / 2}
		dm (groups) = SR0;
		cntr = sr0;		{initialize group counter}
		m4 = sr0;		{twiddle factor modifier}
		m2 = dm (node_space);	{node space modifier}
		i1 = i0;
		modify (i1,m2);		{x1 of the 1-st group}
		i3 = i2; 
		modify (i3,m2);		{y1 of the 1-st group}
	do group_loop until ce;
		i4 = ^twid_real;	{C of W0}
		i5 = ^twid_imag;	{-S of W0}
		cntr = dm (bflys_per_group);
		my0 = pm (i4,m4), mx0 = dm (i1,m0); {my0 = C, mx0 = x1}
		my1 = pm (i5,m4), mx1 = dm (i3,m0); {my1 = -S, mx1 = y1}
		do bfly_loop until ce;
			mr = mx0*my1 (SS), ax0 = dm (i0,m0);
				{mr = x1 (-S), ax0 = x0}
			mr = mr + mx1*my0 (RND), ax1 = dm (i2,m0);
				{mr = y1 (C) + x1 (-S), ax1 = y0}
			ay1 = mr1, mr = mx0*my0 (SS);
				{ay1 = y1 (C) + x1 (-S), mr = x1 (C)}
			mr = mr - mx1*my1 (RND);
				{mr = x1 (C) - y1 (-S)}
			ay0 = mr1, ar = ax1 - ay1;
				{ay0 = x1 (C) - y1 (-S), ar = y0 - [y1 (C) +x1 (-S)]}
			SB = expadj ar, dm (i3,m1) = ar;
			{check for bit growth, y1 = y0 - [y1 (C) + x1 (-S)]}
			ar = ax0 - ay0, mx1 = dm (i3,m0), my1 = pm (i5,m4);
			{ar = x0 - [x1 (C) - y1 (-S)], mx1 = next y1, my1 = next (-S)}
			SB = expadj ar, dm (i1,m1) = ar;
			{check for bit growth, x1 = x0 - [x1 (C) - y1 (-S)]}
			ar = ax0 + ay0, mx0 = dm (i1,m0), my0 = pm (i4,m4);
			{ar = x0 + [x1 (C) - y1 (-S)], mx0 = next x1, my0 = next (C)]}
			SB = expadj ar, dm (i0,m1) =ar;
			{check for bit growth, x0 = x0 + [x1 (C) - y1(-S)]}
			ar = ax1 + ay1;
			{ar = y0 + [y1 (C) + x1 (-S)]}
bfly_loop:		SB = expadj ar, dm(i2,m1) = ar;
			{check for bit growth, y0 = y0 + [y1 (C) + x1 (-S)]}
		modify (i0,m2);		{1-st x0 in next group}
		modify (i1,m2);		{1-st x1 in next group}
		modify (i2,m2);		{1-st y0 in next group}
group_loop:	modify (i3,m2);		{1-st y1 in next group}

		call bfp_adj;
		SI = dm (bflys_per_group);
		SR = ashift SI by 1 (lo);
		dm (node_space) = sr0;	{node_space = 2.node_space}
stage_loop:	dm (bflys_per_group) = sr0;  {bflys_per_group = 2.bflys_per_group}

{**************** LAST STAGE of FFT transform ******************************}
	i0 = ^inplacereal;
	i1 = ^inplacereal + nover2;
	i2 = ^inplaceimag;
	i3 = ^inplaceimag + nover2;
	
	cntr = nover2;
	m2 = dm (node_space);
	m4 = 1;
	i4 = ^twid_real;
	i5 = ^twid_imag;
	
	my0 = pm (i4,m4), mx0 = dm (i1,m0);	{my0 = C, mx0 = x1}
	my1 = pm (i5,m4), mx1 = dm (i3,m0);	{my1 = -S, mx1 = y1}
	do bfly_lp until ce;
		mr = mx0*my1 (SS), ax0 = dm (i0,m0); {mr = x1 (-S), ax0 = x0}
		mr = mr + mx1*my0 (RND), ax1 = dm (i2,m0);
					{mr = y1 (C) + x1 (-S), ax1 = y0}
	ay1 = mr1, mr = mx0*my0 (SS);	{ay1 = y1 (C) + x1 (-S), mr = x1 (C)}
	mr = mr - mx1*my1 (RND);	{mr = x1 (C) - y1 (-S)}
	ay0 = mr1, ar = ax1 - ay1;
			{ay0 = x1 (C) - y1 (-S), ar = y0 - [y1 (C) + x1 (-S)]}
	SB = expadj ar; dm (i3,m1) = ar;
			{check for bit growth, y1 = y0 - [y1 (C) + x1 (-S)]}
	ar = ax0 - ay0, mx1 = dm (i3,m0), my1 = pm (i5,m4);
			{ar = x0 - [x1 (C) - y1 (-S)], mx1 = next y1, my1 = next (-S)}
	SB = expadj ar; dm (i1,m1) = ar;
			{check for bit growth, x1 = x0 - [x1 (C) - y1 (-S)]}
	ar = ax0 + ay0, mx0 = dm (i1,m0), my0 = pm (i4,m4);
			{ar = x0 + [x1 (C) - y1 (-S)], mx0 = next x1, my0 = next C}
	SB = expadj ar, dm (i0,m1) = ar;
			{check for bit growth, x0 = x0 + [x1 (C) - y1 (-S)]}
	ar = ax1 + ay1; 		{ar = y0 + [y1 (C) + x1 (-S)]}
bfly_lp: SB = expadj ar, dm (i2,m1) = ar;
			{check for bit growth, y0 = y0 + [y1 (C) + x1 (-S)]}
	call bfp_adj; 

	call average;	{calls the program, which computes the average}
			{of the spectrum magnitude		      }
	ax0 = N_div_2;
	dm (groups) = ax0;
	ax0 = 2;
	dm (bflys_per_group) = ax0;
	dm (node_space) = ax0;
	ax0 = dm (blk_exponent);
	dm (blockexp) = ax0;	{stores the exponent of the entire block}
	ax0 = 0;
	dm (blk_exponent) = ax0;
	ax0 = 1;
	dm (passed) = ax0;	{sets the FFT passed flag}
jump main;

{************* Block Floating-Point scaling *******************************}
bfp_adj:
	ay0 = cntr;	{check for last stage}
	ar = ay0 - 1;
	{if eq rts;}	{if last stage return}
	ay0 = -2;
	ax0 = SB;
	ar = ax0 - ay0;	{check for SB = -2}
	if eq rts;	{if SB = -2, no bit growth, return}
	i0 = ^inplacereal; {i0=read pointer}
	i1 = ^inplacereal; {i1=write pointer}
	ay0 = -1;
	my0 = H#4000;	{set my0 to shift 1 bit right}
	ar = ax0 - ay0, mx0 = dm (i0,m1);
			{check if SB = -1, get 1-st sample}
	if eq jump strt_shift;
			{if SB = -1, shift block data 1 bit}
	ax0 = -2;	{set ax0 for block exponent update}
	my0 = H#2000;	{set my0 to shift 2 bits right}
strt_shift:
	cntr = Ntimes2 - 1;	{initialize loop counter}
	do shift_loop until ce;	{shift block of data}
		mr =   mx0*my0 (RND), mx0 = dm (i0,m1);
			{mr = shifted data, mx0 = next value}
shift_loop:	dm (i1,m1) = mr1; {unshifted data = shifted data}
	mr = mx0*my0 (RND);	{shift last data word}
	ay0 = dm (blk_exponent);{update block exponent}
	dm (i1,m1) = mr1, ar = ay0 - ax0;
	dm (blk_exponent) = ar;
	rts;



average: 				{This routine takes the average of the}
	i0 = ^inplacereal;		{magnitude of the spectrum. 	      }  
	i1 = ^inplaceimag;		{The resolution is lower, but understanding}
	i3 = ^magnitude;		{is easier, because the spectrum is better }
	m0 = 1;				{vizualized}
	m1 = 1;
	m3 = 1;
	l0 = %inplacereal;
	l1 = %inplaceimag;
	l3 = %magnitude;
	cntr = l0;
	ar = pass 0;{-1;}		{first compute the magnitude}

	do magn until ce;
		mx0 = dm (i0,m0);	{reads the data from the inplacereal buffer}
		mx1 = dm (i1,m1);	{reads the data from the inplaceimag buffer}
		mr = mx0*mx0 (ss);
		mr = mr + mx1*mx1 (rnd); {computes the amplitude of the spectrum}
	magn:	dm (i3,m3) = mr1;
	
	cntr = l0;			{This part of the routine computes    }
	i1 = ^magnitude;		{the square root of the sum of squares}
	m1 = 1;				{of the real and imaginary part of the}
	l1 = %magnitude;		{spectrum			      }	
	m2 = 0;
	ax0 = ^squareval;		{This is the base pointer of the      }
	ay0 = dm (i3,m3);		{look-up table			      }	
	do squareroot until ce;
		ar = ax0 + ay0, ay0 = dm (i3,m3);
		i2 = ar;		{The new value in the look-up is at   }
		ax1 = dm (i2,m2);	{the address, specified by i2	      }
	squareroot:	dm (i1,m1) = ax1;
		

	i0 = ^magnitude;
	i1 = ^magnitude;
	i2 = ^magnitude;
	i3 = ^magnitude;
	i4 = ^magnitude;
	i5 = ^averageval;
	m0 = 1;
	m1 = 1;
	m2 = 2;
	m3 = 3;
	m5 = 1;
	l0 = %magnitude;
	l1 = %magnitude;
	l2 = %magnitude;
	l3 = %magnitude;
	l5 = %averageval;
	modify (i1,m1);
	modify (i2,m2);
	modify (i3,m3);
	m2 = 1;
	m3 = 1;
	cntr = l0;
	do avg until ce;		{compute the average of 4 samples}
		ar = 0;
		ay0 = dm (i0,m0);
		ar = ar + ay0, ay0 = dm (i1,m1);
		ar = ar + ay0, ay0 = dm (i2,m2);
		ar = ar + ay0; ay0 = dm (i3,m3);
		ar = ar + ay0; 
		sr = ashift ar by -2 (hi);
	avg:	dm (i5,m5) = sr1; 
	rts;

	

{****** Interrupt service routines ******************************}

{****** Receive interrupt service routine ***********************}
input_samples:
	ena sec_reg;		{use shadow register bank}

	ax1 = dm (rx_buf + 1);	{get data from CODEC}
	{si = ax1;}		{eventually attenuate the input}
	{sr = ashift si by -2 (hi);}
	dm (i0,m0) = ax1;	{writes the data to the inputreal buffer}
	my1 = dm (rx_buf + 2);

	ax0 = i0;		{test if the whole data buffer is full}
	ay0 = ^inputreal;
	ar = ax0 - ay0;
	if eq ar = pass 0;	
	dm (passed) = ar;

	dm (tx_buf + 1) = ax1;	{send data to CODEC}
	dm (tx_buf + 2) = my1;

	ay0 = dm (blockexp);
	ar = -ay0;
	ar = ar - 9;		{restores the original exponent}
	ar = ar - 3;
	se = ar;

	ax0 = dm (i1,m1);
	si = ax0;		{maximum resolution spectrum}
	sr = norm si (lo);
	IO(DACA) = sr1;		{sends data to DAC-8412 DACA}

	ax0 = dm (i3,m3);
	si = ax0;		{averaged spectrum}
	sr = norm si (lo);
	IO(DACB) = sr1;		{sends averaged data to DAC-8412 DACB}
	dis sec_reg;
	rti;

{****** Transmit interrupt routine used for CODEC initialization ******}
next_cmd:
	ena sec_reg;
	ax0 = dm (i3, m1);	{fetch next control word}
	dm (tx_buf) = ax0;	{place in transmit slot 0}
	ax0 = i3;
	ay0 = ^init_cmds;
	ar = ax0 - ay0;
	if gt rti;		{rti if more control words}
	ax0 = 0x8000;		{else set done flag}
	dm (tx_buf) = ax0;	{remove MCE}
	ax0 = 0;
	dm (stat_flag) = ax0;	{reset status flag}
	rti;

	
.endmod;
